#ifndef _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NVC0_M2MF_XML
#define _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NVC0_M2MF_XML

/* Autogenerated file, DO NOT EDIT manually!

This file was generated by the rules-ng-ng headergen tool in this git repository:
http://0x04.net/cgit/index.cgi/rules-ng-ng
git clone git://0x04.net/rules-ng-ng

The rules-ng-ng source files this header was generated from are:
- /home/skeggsb/git/envytools/rnndb/nv_objects.xml      (    794 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/copyright.xml       (   6452 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_m2mf.xml         (   2696 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_object.xml       (  12672 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvchipsets.xml      (   3617 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_defs.xml         (   4437 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_defs.xml       (   5468 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvc0_m2mf.xml       (   2687 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv01_2d.xml         (  32584 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv04_dvd.xml        (   3000 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv03_3d.xml         (   5209 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv04_3d.xml         (  17759 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv_3ddefs.xml       (  16394 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv10_3d.xml         (  18437 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv20_3d.xml         (  21107 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv30-40_3d.xml      (  31987 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_2d.xml         (  11113 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv50_3d.xml         (  65233 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nv50_compute.xml    (  14012 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nv84_crypt.xml      (   2071 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nv31_mpeg.xml       (   2269 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/nvc0_3d.xml         (  52547 bytes, from 2011-11-30 05:49:35)
- /home/skeggsb/git/envytools/rnndb/nvc0_compute.xml    (  10865 bytes, from 2011-10-22 08:01:09)
- /home/skeggsb/git/envytools/rnndb/blob_nvc0_pcopy.xml (   4516 bytes, from 2011-10-22 08:01:09)

Copyright (C) 2006-2011 by the following authors:
- Artur Huillet <arthur.huillet@free.fr> (ahuillet)
- Ben Skeggs (darktama, darktama_)
- B. R. <koala_br@users.sourceforge.net> (koala_br)
- Carlos Martin <carlosmn@users.sf.net> (carlosmn)
- Christoph Bumiller <e0425955@student.tuwien.ac.at> (calim, chrisbmr)
- Dawid Gajownik <gajownik@users.sf.net> (gajownik)
- Dmitry Baryshkov
- Dmitry Eremin-Solenikov <lumag@users.sf.net> (lumag)
- EdB <edb_@users.sf.net> (edb_)
- Erik Waling <erikwailing@users.sf.net> (erikwaling)
- Francisco Jerez <currojerez@riseup.net> (curro)
- imirkin <imirkin@users.sf.net> (imirkin)
- jb17bsome <jb17bsome@bellsouth.net> (jb17bsome)
- Jeremy Kolb <kjeremy@users.sf.net> (kjeremy)
- Laurent Carlier <lordheavym@gmail.com> (lordheavy)
- Luca Barbieri <luca@luca-barbieri.com> (lb, lb1)
- Maarten Maathuis <madman2003@gmail.com> (stillunknown)
- Marcin Kościelnicki <koriakin@0x04.net> (mwk, koriakin)
- Mark Carey <mark.carey@gmail.com> (careym)
- Matthieu Castet <matthieu.castet@parrot.com> (mat-c)
- nvidiaman <nvidiaman@users.sf.net> (nvidiaman)
- Patrice Mandin <patmandin@gmail.com> (pmandin, pmdata)
- Pekka Paalanen <pq@iki.fi> (pq, ppaalanen)
- Peter Popov <ironpeter@users.sf.net> (ironpeter)
- Richard Hughes <hughsient@users.sf.net> (hughsient)
- Rudi Cilibrasi <cilibrar@users.sf.net> (cilibrar)
- Serge Martin
- Simon Raffeiner
- Stephane Loeuillet <leroutier@users.sf.net> (leroutier)
- Stephane Marchesin <stephane.marchesin@gmail.com> (marcheu)
- sturmflut <sturmflut@users.sf.net> (sturmflut)
- Sylvain Munaut <tnt@246tNt.com>
- Victor Stinner <victor.stinner@haypocalc.com> (haypo)
- Wladmir van der Laan <laanwj@gmail.com> (miathan6)
- Younes Manton <younes.m@gmail.com> (ymanton)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE COPYRIGHT OWNER(S) AND/OR ITS SUPPLIERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/



#define NVC0_M2MF_TILING_MODE_IN				0x00000204

#define NVC0_M2MF_TILING_PITCH_IN				0x00000208

#define NVC0_M2MF_TILING_HEIGHT_IN				0x0000020c

#define NVC0_M2MF_TILING_DEPTH_IN				0x00000210

#define NVC0_M2MF_TILING_POSITION_IN_Z				0x00000214

#define NVC0_M2MF_TILING_MODE_OUT				0x00000220

#define NVC0_M2MF_TILING_PITCH_OUT				0x00000224

#define NVC0_M2MF_TILING_HEIGHT_OUT				0x00000228

#define NVC0_M2MF_TILING_DEPTH_OUT				0x0000022c

#define NVC0_M2MF_TILING_POSITION_OUT_Z				0x00000230

#define NVC0_M2MF_OFFSET_OUT_HIGH				0x00000238

#define NVC0_M2MF_OFFSET_OUT_LOW				0x0000023c

#define NVC0_M2MF_FIRMWARE(i0)				       (0x0000027c + 0x4*(i0))
#define NVC0_M2MF_FIRMWARE__ESIZE				0x00000004
#define NVC0_M2MF_FIRMWARE__LEN					0x00000020

#define NVC0_M2MF_EXEC						0x00000300
#define NVC0_M2MF_EXEC_PUSH					0x00000001
#define NVC0_M2MF_EXEC_LINEAR_IN				0x00000010
#define NVC0_M2MF_EXEC_LINEAR_OUT				0x00000100
#define NVC0_M2MF_EXEC_QUERY__MASK				0x00003000
#define NVC0_M2MF_EXEC_QUERY__SHIFT				12
#define NVC0_M2MF_EXEC_QUERY_NO					0x00000000
#define NVC0_M2MF_EXEC_QUERY_UNK1				0x00001000
#define NVC0_M2MF_EXEC_QUERY_YES				0x00002000
#define NVC0_M2MF_EXEC_BUF_NOTIFY				0x00030000
#define NVC0_M2MF_EXEC_QUERY_SHORT				0x00100000

#define NVC0_M2MF_DATA						0x00000304

#define NVC0_M2MF_UNK308					0x00000308

#define NVC0_M2MF_OFFSET_IN_HIGH				0x0000030c

#define NVC0_M2MF_OFFSET_IN_LOW					0x00000310

#define NVC0_M2MF_PITCH_IN					0x00000314

#define NVC0_M2MF_PITCH_OUT					0x00000318

#define NVC0_M2MF_LINE_LENGTH_IN				0x0000031c

#define NVC0_M2MF_LINE_COUNT					0x00000320

#define NVC0_M2MF_QUERY_ADDRESS_HIGH				0x0000032c

#define NVC0_M2MF_QUERY_ADDRESS_LOW				0x00000330

#define NVC0_M2MF_QUERY_SEQUENCE				0x00000334

#define NVC0_M2MF_COND_ADDRESS_HIGH				0x00000338

#define NVC0_M2MF_COND_ADDRESS_LOW				0x0000033c

#define NVC0_M2MF_COND_MODE					0x00000340
#define NVC0_M2MF_COND_MODE_NEVER				0x00000000
#define NVC0_M2MF_COND_MODE_ALWAYS				0x00000001
#define NVC0_M2MF_COND_MODE_RES_NON_ZERO			0x00000002
#define NVC0_M2MF_COND_MODE_EQUAL				0x00000003
#define NVC0_M2MF_COND_MODE_NOT_EQUAL				0x00000004

#define NVC0_M2MF_TILING_POSITION_IN_X				0x00000344

#define NVC0_M2MF_TILING_POSITION_IN_Y				0x00000348

#define NVC0_M2MF_TILING_POSITION_OUT_X				0x0000034c

#define NVC0_M2MF_TILING_POSITION_OUT_Y				0x00000350


#endif /* _HOME_SKEGGSB_GIT_ENVYTOOLS_RNNDB_NVC0_M2MF_XML */
